def mergeSeriesOnTimestampRange(df_cpu, df_type):
  """
  Upsamples df_type to df_cpu.
  Example:
    Input
      df_cpu = pd.Series({'time': [1,2,3,4], 'field_1': [5,6,7,8]})
      df_type = pd.Series({'time': [1,3], 'field_2': ['a','b']})
    Returns
      pd.Series({'time': [1,2,3,4], 'field_1': [5,6,7,8], 'field_2': ['a','a','b','b']})
  """
  import numpy as np

  df_cpu['instanceType'] = None
  # assume df_type is sorted in decreasing EventTime order (very important)
  # NB: since some instances are not present in the cloudtrail (for which we append artificially the "now" type)
  #     Need to traverse the df_type matrix backwards
  for index, row_type in df_type.iterrows():
      # use row_type.name instead of row_type['EventTime']
      # check note above about needing to traverse backwards
      # df_cpu.iloc[np.where(df_cpu.Timestamp >= row_type.name)[0], df_cpu.columns.get_loc('instanceType')] = row_type['instanceType']
      df_cpu.iloc[np.where(df_cpu.Timestamp <= row_type.name)[0], df_cpu.columns.get_loc('instanceType')] = row_type['instanceType']

  # fill na at beginning with back-fill
  # (artifact of cloudwatch having data at days before the creation of the instance)
  df_cpu['instanceType'] = df_cpu['instanceType'].fillna(method='backfill')
  return df_cpu







def ec2_catalog():
    import requests
    from cachecontrol import CacheControl
    from cachecontrol.caches.file_cache import FileCache

    import logging
    logger = logging.getLogger('isitfit')
    logger.debug("Downloading ec2 catalog (cached to local file)")

    # based on URL = 'http://www.ec2instances.info/instances.json'
    # URL = 's3://...csv'
    # Edit 2019-09-10 use CDN link instead of direct gitlab link
    # URL = 'https://gitlab.com/autofitcloud/www.ec2instances.info-ec2op/raw/master/www.ec2instances.info/t3b_smaller_familyL2.json'
    URL = 'https://cdn.jsdelivr.net/gh/autofitcloud/www.ec2instances.info-ec2op/www.ec2instances.info/t3b_smaller_familyL2.json'

    # cached https://cachecontrol.readthedocs.io/en/latest/
    sess = requests.session()
    cached_sess = CacheControl(sess, cache=FileCache('/tmp/isitfit_ec2info.cache'))
    r = cached_sess.request('get', URL)

    # read catalog, copy from ec2op-cli/ec2op/optimizer/cwDailyMaxMaxCpu
    import json
    j = json.dumps(r.json(), indent=4, sort_keys=True)
    from pandas import read_json
    df = read_json(j, orient='split')
    
    # Edit 2019-09-13 no need to subsample the columns at this stage
    # df = df[['API Name', 'Linux On Demand cost']]

    df = df.rename(columns={'Linux On Demand cost': 'cost_hourly'})
    # df = df.set_index('API Name') # need to use merge, not index
    return df


# copied from git-remote-aws
def mysetlocale():
  li = 'en_US.utf8'
  import os
  os.environ["LC_ALL"] = li
  os.environ["LANG"]   = li




MAX_ROWS = 10
MAX_COLS = 5
MAX_STRING = 20
def display_df(title, df, csv_fn, shape, logger):
    # https://pypi.org/project/termcolor/
    from termcolor import colored

    logger.info("")

    if shape[0]==0:
      logger.info(title)
      logger.info(colored("None", "red"))
      return

    if csv_fn is not None:
      logger.info(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
      logger.info(colored("It could be opened in the terminal with visidata (http://visidata.org/)","cyan"))
      logger.info(colored("and you can close visidata by pressing 'q'","cyan"))
      open_vd = input(colored('Would you like to do so? yes/[no] ', 'cyan'))
      if open_vd.lower() == 'yes' or open_vd.lower() == 'y':
        logger.info("Opening CSV file `%s` with visidata."%csv_fn)
        from subprocess import call
        call(["vd", csv_fn])
        logger.info("Exited visidata.")
        logger.info(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
        return
      else:
        logger.info("Not opening visidata.")
        logger.info("To open the results with visidata, use `vd %s`."%csv_fn)


    # if not requested to open with visidata
    from tabulate import tabulate
    df_show = df.head(n=MAX_ROWS)
    df_show = df_show.applymap(lambda c: (c[:MAX_STRING]+'...' if len(c)>=MAX_STRING else c) if type(c)==str else c)

    logger.info(tabulate(df_show, headers='keys', tablefmt='psql', showindex=False))

    if (shape[0] > MAX_ROWS) or (shape[1] > MAX_COLS):
      logger.info("...")
      logger.info("(results truncated)")
      # done
      return

    # done
    return


class IsitfitError(Exception):
  pass



from .datadogManager import DatadogManager

def test_datadogman_1():
    ddg = DatadogManager()
    host_id='i-0f31feed76f7fb07c'
    df_all = ddg.get_metrics_all(host_id=host_id)
    assert df_all.shape[0] > 0
    #import pdb
    #pdb.set_trace()
    #print(df_all.head())




from .tagsPush import TagsPush
import pytest
from .utils import IsitfitError

def test_validateTagsFile_fail_empty():
  import tempfile
  with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-testTagsPush-', delete=True) as fh:
    tp = TagsPush(fh.name)
    with pytest.raises(IsitfitError) as e_info:
      tp.read_csv()
      # tp.validateTagsFile()


def test_validateTagsFile_fail_format_1():
  import pandas as pd
  df = pd.DataFrame([
    {'A': 'abc', 'B': 'some instance'},
    {'A': 'def', 'B': 'some instance'},
  ])
  import tempfile
  with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-testTagsPush-', delete=True) as fh:
    df.to_csv(fh.name, index=False)
    tp = TagsPush(fh.name)
    with pytest.raises(IsitfitError) as e_info:
      tp.read_csv()


def test_validateTagsFile_fail_format_2():
  import pandas as pd
  df = pd.DataFrame([
    {'instance_id': 'i-1', 'A': 'abc', 'B': 'some instance'},
    {'instance_id': 'i-2', 'A': 'def', 'B': 'some instance'},
  ])
  import tempfile
  with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-testTagsPush-', delete=True) as fh:
    df.to_csv(fh.name, index=False)
    tp = TagsPush(fh.name)
    tp.read_csv()
    with pytest.raises(IsitfitError) as e_info:
      tp.validateTagsFile()


def test_validateTagsFile_ok():
  import pandas as pd
  df = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance', 'Environment': 'Dev'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Environment': 'Dev'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Environment': 'Dev'},
  ])
  import tempfile
  with tempfile.NamedTemporaryFile(suffix='.csv', prefix='isitfit-testTagsPush-', delete=True) as fh:
    df.to_csv(fh.name, index=False)
    tp = TagsPush(fh.name)
    tp.read_csv()
    tp.validateTagsFile()

#-----------------------

def test_diffLatest_fail_noChanges(monkeypatch):
  """
  No changes detected
  """
  import pandas as pd
  tags_old = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance', 'Environment': 'Dev'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Environment': 'Dev'},
  ])
  tags_new = tags_old.copy()

  # perform diff
  tp = TagsPush(None)
  tp.csv_df = tags_new
  tp.latest_df = tags_old
  with pytest.raises(IsitfitError) as e_info:
    tp.diffLatest()



def test_diffLatest_fail_newInstances(monkeypatch):
  """
  Some instance IDs in new but not in old
  """
  import pandas as pd
  tags_old = pd.DataFrame([
    {'instance_id': 'i-1'},
    {'instance_id': 'i-2'},
  ])
  tags_new = pd.DataFrame([
    {'instance_id': 'i-1'},
    {'instance_id': 'i-3'},
  ])

  # perform diff
  tp = TagsPush(None)
  tp.csv_df = tags_new
  tp.latest_df = tags_old
  with pytest.raises(IsitfitError) as e_info:
    tp.diffLatest()


def test_diffLatest_fail_dropIsWrong(monkeypatch):
  """
  Environment tag is dropped but user disagrees
  """
  import pandas as pd
  tags_old = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance', 'Environment': 'Dev'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Environment': 'Dev'},
  ])
  tags_new = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance'},
    {'instance_id': 'i-2', 'Name': 'another instance'},
  ])

  # How to test a function with input call?
  # https://stackoverflow.com/a/36377194/4126114
  # monkeypatch the "input" function, so that it returns "Mark".
  # This simulates the user entering "Mark" in the terminal:
  monkeypatch.setattr('builtins.input', lambda x: "no")

  # perform diff
  tp = TagsPush(None)
  tp.csv_df = tags_new
  tp.latest_df = tags_old
  with pytest.raises(IsitfitError) as e_info:
    tp.diffLatest()



def test_diffLatest_ok(monkeypatch):
  """
  Environment tag is dropped but user disagrees
  """
  import pandas as pd
  tags_old = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance', 'Environment': 'Dev', 'App': 'db'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Environment': 'Dev', 'App': 'db'},
  ])
  tags_new = pd.DataFrame([
    {'instance_id': 'i-1', 'Name': 'some instance', 'Owner': 'shadi', 'Application': 'db'},
    {'instance_id': 'i-2', 'Name': 'another instance', 'Owner': 'tarzan', 'Application': 'db'},
  ])

  # How to test a function with input call?
  # https://stackoverflow.com/a/36377194/4126114
  # monkeypatch the "input" function, so that it returns "Mark".
  # This simulates the user entering "Mark" in the terminal:
  monkeypatch.setattr('builtins.input', lambda x: "yes")

  # perform diff
  tp = TagsPush(None)
  tp.csv_df = tags_new
  tp.latest_df = tags_old
  tp.diffLatest()
  actual_mig = tp.mig_df

  expected_mig = pd.DataFrame([
    {'action': 'mv',    'old': 'App',         'new': 'Application'},
    {'action': 'touch', 'old': None,          'new': 'Owner'},
    {'action': 'rm',    'old': 'Environment', 'new': None},
  ])
  pd.testing.assert_frame_equal(actual_mig, expected_mig)



"""
Manager class for getting datadog data for isitfit

Pre-requisites
    pip3 install datadog asciiplotlib pandas
    apt-get install gnuplot

set env vars
    export DATADOG_API_KEY=...
    export DATADOG_APP_KEY=...
    
Run tests
    pip3 install pytest
    pytest datadogManager.py
"""

import json
import time
import pandas as pd
import os

import logging
logger = logging.getLogger('isitfit')

# Configure the module according to your needs
from datadog import initialize

# Use Datadog REST API client
from datadog import api

SECONDS_IN_ONE_DAY = 60*60*24

class HostNotFoundInDdg(ValueError):
  pass

class DataNotFoundForHostInDdg(ValueError):
  pass

class DatadogAssistant:
    def __init__(self, start, end, host_id):
        self.end = end
        self.start = start
        self.host_id = host_id

    def _get_metrics_core(self, query, col_i):
        m = api.Metric.query(start=self.start, end=self.end, query=query, host=self.host_id)
        if len(m['series'])==0:
            raise DataNotFoundForHostInDdg("No %s found for %s"%(col_i, self.host_id))
        df = pd.DataFrame(m['series'][0]['pointlist'], columns=['ts_int', col_i])
        #import pdb
        #pdb.set_trace()
        df['ts_dt'] = pd.to_datetime(df.ts_int, origin='unix', unit='ms')
        return df
        
    def _get_meta(self):
        h_all = api.Hosts.search(host=self.host_id)
        if len(h_all['host_list'])==0:
            raise HostNotFoundInDdg("Did not find host %s in datadog"%self.host_id)

        h_i = h_all['host_list'][0]
        gohai = json.loads(h_i['meta']['gohai'])
        memory_total = int(gohai['memory']['total'].replace('kB',''))*1024
        out = {'cpuCores': h_i['meta']['cpuCores'], 'memory_total': memory_total}
        return out
        
    def get_metrics_cpu_max(self):
        # query language
        # https://docs.datadoghq.com/graphing/functions/
        # Use minimum so that cpu_used will be the maximum
        query = 'system.cpu.idle{host:%s}.rollup(min,%i)'%(self.host_id, SECONDS_IN_ONE_DAY)
        col_i = 'cpu_idle_min'
        df = self._get_metrics_core(query, col_i)
        # calculate cpu used as 100 - cpu_idle
        df['cpu_used_max'] = 100 - df.cpu_idle_min
        df['cpu_used_max'] = df['cpu_used_max'].astype(int)
        return df
        
    def get_metrics_cpu_avg(self):
        # repeat for average
        query = 'system.cpu.idle{host:%s}.rollup(avg,%i)'%(self.host_id, SECONDS_IN_ONE_DAY)
        col_i = 'cpu_idle_avg'
        df = self._get_metrics_core(query, col_i)
        df['cpu_used_avg'] = 100 - df.cpu_idle_avg
        df['cpu_used_avg'] = df['cpu_used_avg'].astype(int)
        return df

    def get_metrics_ram_max(self):
        # query language, check note above in get_metrics_cpu
        query = 'system.mem.free{host:%s}.rollup(min,%i)'%(self.host_id, SECONDS_IN_ONE_DAY)
        col_i = 'ram_free_min'
        df =  self._get_metrics_core(query, col_i)
        memory_total = self._get_meta()['memory_total']
        df['ram_free_min'] = df.ram_free_min / memory_total * 100
        df['ram_free_min'] = df['ram_free_min'].astype(int)
        df['ram_used_max'] = 100 - df['ram_free_min']
        return df

    def get_metrics_ram_avg(self):
        # query language, check note above in get_metrics_cpu
        query = 'system.mem.free{host:%s}.rollup(avg,%i)'%(self.host_id, SECONDS_IN_ONE_DAY)
        col_i = 'ram_free_avg'
        df =  self._get_metrics_core(query, col_i)
        memory_total = self._get_meta()['memory_total']
        df['ram_free_avg'] = df.ram_free_avg / memory_total * 100
        df['ram_free_avg'] = df['ram_free_avg'].astype(int)
        df['ram_used_avg'] = 100 - df['ram_free_avg']
        return df

class DatadogManager:
    
    def __init__(self):
        initialize()
        self.end = int(time.time())
        
        # datadog will automatically set the resolution based on the start-end range
        # if the ".rollup" is not used in the query
        # number of seconds in 1 hour, if no ".rollup" specified in query, this yields 20-second frequency from datadog
        # n_secs = 60*60
        # number of seconds in 90 days, if no ".rollup" specified in query, yields bi-daily from datadog
        n_secs = SECONDS_IN_ONE_DAY*90
        self.start = self.end - n_secs


    def is_configured(self):
      if os.getenv('DATADOG_API_KEY', None) is not None:
        if os.getenv('DATADOG_APP_KEY', None) is not None:
          return True
          
      return False


    def get_metrics_all(self, host_id):
        # FIXME: we already have cpu from cloudwatch, so maybe just focus on ram from datadog
        logger.debug("Fetching datadog data for %s"%host_id)
        ddgL2 = DatadogAssistant(self.start, self.end, host_id)
        df_cpu_max = ddgL2.get_metrics_cpu_max()
        df_cpu_avg = ddgL2.get_metrics_cpu_avg()
        df_ram_max = ddgL2.get_metrics_ram_max()
        df_ram_avg = ddgL2.get_metrics_ram_avg()
        df_all = (
            df_cpu_max
            .merge(df_cpu_avg, how='outer', on=['ts_int', 'ts_dt'])
            .merge(df_ram_max, how='outer', on=['ts_int', 'ts_dt'])
            .merge(df_ram_avg, how='outer', on=['ts_int', 'ts_dt'])
        )
        df_all = df_all[['ts_dt', 'cpu_used_max', 'cpu_used_avg', 'ram_used_max', 'ram_used_avg']]
        return df_all



import logging
logger = logging.getLogger('isitfit')

import pandas as pd
from tabulate import tabulate
import tempfile
import csv
from collections import OrderedDict

# https://pypi.org/project/termcolor/
from termcolor import colored


def df2tabulate(df):
  return tabulate(df.set_index('instance_id'), headers='keys', tablefmt='psql')


def class2recommendedType(r):
  if r.classification_1 == 'Underused':
    # FIXME classification 2 will contain if it's a burstable workload or lambda-convertible
    # that would mean that the instance is downsizable twice, so maybe need to return r.type_smaller2x
    return r.type_smaller

  if r.classification_1=='Overused':
    return r.type_larger

  return None


def class2recommendedCost(r):
  if r.classification_1 == 'Underused':
    # FIXME add savings from the twice downsizing in class2recommendedType if it's a burstable workload or lambda-convertible,
    # then calculate the cost from lambda functions and add it as overhead here
    return r.cost_3m_smaller-r.cost_3m

  if r.classification_1=='Overused':
    return r.cost_3m_larger-r.cost_3m

  return None


def ec2obj_to_name(ec2_obj):
    if ec2_obj.tags is None:
      return None

    ec2_name = [x for x in ec2_obj.tags if x['Key']=='Name']
    if len(ec2_name)==0:
      return None

    return ec2_name[0]['Value']


class OptimizerListener:

  def __init__(self, n, thresholds = None):
    self.n = n

    if thresholds is None:
      thresholds = {
        'rightsize': {'idle': 3, 'low': 30, 'high': 70},
        'burst': {'low': 20, 'high': 80}
      }

    # iterate over all ec2 instances
    self.thresholds = thresholds
    self.ec2_classes = []

    # members that will contain the results of the optimization
    self.df_sort = None
    self.sum_val = None

    # for csv streaming
    self.csv_fn_intermediate = None
    self.csv_fn_final = None
    self.csv_fh = None
    self.csv_writer = None

  
  def __exit__(self):
    self.csv_fh.close()


  def _xxx_to_classification(self, xxx_maxmax, xxx_maxavg, xxx_avgmax):
    # check if good to convert to burstable or lambda
    # i.e. daily data shows few large spikes
    thres = self.thresholds['burst']
    if xxx_maxmax >= thres['high'] and xxx_avgmax <= thres['low'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable daily'

    # check rightsizing
    # i.e. no special spikes in daily data
    # FIXME: can check hourly data for higher precision here
    thres = self.thresholds['rightsize']
    if xxx_maxmax <= thres['idle']:
      return 'Idle', None
    elif xxx_maxmax <= thres['low']:
      return 'Underused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg >= thres['high']:
      return 'Overused', None
    elif xxx_maxmax >= thres['high'] and xxx_avgmax >= thres['high'] and xxx_maxavg <= thres['low']:
      return 'Underused', 'Burstable intraday'

    return 'Normal', None


  def _ec2df_to_classification(self, ec2_df, ddg_df):
    cpu_maxmax = ec2_df.Maximum.max()
    cpu_maxavg = ec2_df.Average.max()
    cpu_avgmax = ec2_df.Maximum.mean()
    cpu_c1, cpu_c2 = self._xxx_to_classification(cpu_maxmax, cpu_maxavg, cpu_avgmax)
    #print("ec2_df.{maxmax,avgmax,maxavg} = ", maxmax, avgmax, maxavg)

    if ddg_df is None:
      cpu_c2 = [cpu_c2, "No memory data"]
      cpu_c2 = [x for x in cpu_c2 if x is not None]
      cpu_c2 = ", ".join(cpu_c2)
      return cpu_c1, cpu_c2

    # continue with datadog data
    ram_maxmax = ddg_df.ram_used_max.max()
    ram_maxavg = ddg_df.ram_used_max.mean()
    ram_avgmax = ddg_df.ram_used_avg.max()
    ram_c1, ram_c2 = self._xxx_to_classification(ram_maxmax, ram_maxavg, ram_avgmax)

    # consolidate ram with cpu
    out_c2 = ["CPU + RAM checked",
              "CPU: %s"%cpu_c2 if cpu_c2 is not None else None,
              "RAM: %s"%ram_c2 if ram_c2 is not None else None ]
    out_c2 = ", ".join([x for x in out_c2 if x is not None])
    if cpu_c1=='Overused' or ram_c1=='Overused':
      return 'Overused', out_c2

    if cpu_c1=='Normal' or ram_c1=='Normal':
      return 'Normal', out_c2

    return 'Underused', out_c2


  def handle_pre(self):
      # a csv file handle to which to stream results
      self.csv_fn_intermediate = tempfile.NamedTemporaryFile(prefix='isitfit-intermediate-', suffix='.csv', delete=False)
      logger.info(colored("Intermediate results will be streamed to %s"%self.csv_fn_intermediate.name, "cyan"))
      self.csv_fh = open(self.csv_fn_intermediate.name, 'w')
      self.csv_writer = csv.writer(self.csv_fh)


  def per_ec2(self, ec2_obj, ec2_df, mm, ddg_df):
    #print(ec2_obj.instance_id)
    ec2_c1, ec2_c2 = self._ec2df_to_classification(ec2_df, ddg_df)

    ec2_name = ec2obj_to_name(ec2_obj)

    taglist = ec2_obj.tags
    if mm.filter_tags is not None:
      f_tn = mm.filter_tags.lower()
      # similar to the isitfit.mainManager.tagsContain function, but filtering the tags themselves
      taglist = [x for x in taglist if (f_tn in x['Key'].lower()) or (f_tn in x['Value'].lower())]

    taglist = ["%s = %s"%(x['Key'], x['Value']) for x in taglist]
    taglist = "\n".join(taglist)

    ec2_res = OrderedDict()
    ec2_res['instance_id'] = ec2_obj.instance_id
    ec2_res['instance_type'] = ec2_obj.instance_type
    ec2_res['classification_1'] = ec2_c1
    ec2_res['classification_2'] = ec2_c2
    ec2_res['tags'] = taglist

    # write csv header
    if len(self.ec2_classes)==0:
      self.csv_writer.writerow(['name'] + [k for k,v in ec2_res.items() if k!='tags'])# save header

    # save intermediate result to csv file
    # Try to stick to 1 row per instance
    # Drop the tags because they're too much to include
    csv_row = [ec2_name] + [v.replace("\n", ";") for k,v in ec2_res.items() if k!='tags']
    self.csv_writer.writerow(csv_row)

    # gathering results
    self.ec2_classes.append(ec2_res)

    # check if should return early
    if self.n!=0:
      sub_underused = [x for x in self.ec2_classes if x['classification_1']=='Underused']
      if len(sub_underused) >= self.n:
        # break early
        self.after_all(None, mm, None)
        import sys
        sys.exit(0)


  def after_all(self, n_ec2_total, mm, n_ec2_all):
    df_all = pd.DataFrame(self.ec2_classes)

    # if no data
    if df_all.shape[0]==0:
      self.df_sort = None
      self.sum_val = None
      return

    # merge current type hourly cost
    map_cost = mm.df_cat[['API Name', 'cost_hourly']]
    df_all = df_all.merge(map_cost, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge the next-smaller instance type from the catalog for instances classified as Underused
    map_smaller = mm.df_cat[['API Name', 'type_smaller', 'Linux On Demand cost_smaller']].rename(columns={'Linux On Demand cost_smaller': 'cost_hourly_smaller'})
    df_all = df_all.merge(map_smaller, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # merge next-larger instance type
    map_larger = mm.df_cat[['API Name', 'type_smaller', 'cost_hourly']].rename(columns={'type_smaller': 'API Name', 'API Name': 'type_larger', 'cost_hourly': 'cost_hourly_larger'})
    df_all = df_all.merge(map_larger, left_on='instance_type', right_on='API Name', how='left').drop(['API Name'], axis=1)

    # convert from hourly to 3-months
    for fx1, fx2 in [('cost_3m', 'cost_hourly'), ('cost_3m_smaller', 'cost_hourly_smaller'), ('cost_3m_larger', 'cost_hourly_larger')]:
      df_all[fx1] = df_all[fx2] * 24 * 30 * 3
      df_all[fx1] = df_all[fx1].fillna(value=0).astype(int)

    # imply a recommended type
    df_all['recommended_type'] = df_all.apply(class2recommendedType, axis=1)
    df_all['savings'] = df_all.apply(class2recommendedCost, axis=1)
    df_all['savings'] = df_all.savings.fillna(value=0).astype(int)

    # keep a subset of columns
    df_all = df_all[['instance_id', 'instance_type', 'classification_1', 'classification_2', 'cost_3m', 'recommended_type', 'savings', 'tags']]

    # display
    #df_all = df_all.set_index('classification_1')
    #for v in ['Idle', 'Underused', 'Overused', 'Normal']:
    #  logger.info("\nInstance classification_1: %s"%v)
    #  if v not in df_all.index:
    #    logger.info("None")
    #  else:
    #    logger.info(df_all.loc[[v]]) # use double brackets to maintain single-row dataframes https://stackoverflow.com/a/45990057/4126114
    #
    #  logger.info("\n")

    # main results
    self.df_sort = df_all.sort_values(['savings'], ascending=True)
    self.sum_val = df_all.savings.sum()


  def storecsv_all(self, *args, **kwargs):
      if self.df_sort is None:
        return

      with tempfile.NamedTemporaryFile(prefix='isitfit-full-', suffix='.csv', delete=False) as  csv_fh_final:
        self.csv_fn_final = csv_fh_final.name
        logger.debug(colored("Saving final results to %s"%csv_fh_final.name, "cyan"))
        self.df_sort.to_csv(csv_fh_final.name, index=False)
        logger.debug(colored("Save complete", "cyan"))


  def display_all(self, *args, **kwargs):
    if self.df_sort is None:
      logger.info(colored("No EC2 instances found", "red"))
      return

    # display
    # Edit 2019-09-25 just show the full list. Will add filtering later. This way it's less ambiguous when all instances are "Normal"
    # self.df_sort.dropna(subset=['recommended_type'], inplace=True)
    
    # if no recommendations
    if self.df_sort.shape[0]==0:
      logger.info(colored("No optimizations from isitfit for this AWS EC2 account", "red"))
      return
    
    # if there are recommendations, show them
    sum_comment = "extra cost" if self.sum_val>0 else "savings"
    sum_color = "red" if self.sum_val>0 else "green"

    #logger.info("Optimization based on the following CPU thresholds:")
    #logger.info(self.thresholds)
    #logger.info("")
    logger.info(colored("Recommended %s: %0.0f $ (over the next 3 months)"%(sum_comment, self.sum_val), sum_color))
    logger.info("")

    # display dataframe
    from .utils import display_df
    display_df(
      "Recommended EC2 size changes",
      self.df_sort,
      self.csv_fn_final,
      self.df_sort.shape,
      logger
    )

#    with pd.option_context("display.max_columns", 10):
#      logger.info("Details")
#      if self.df_sort.shape[0]<=10:
#        logger.info(df2tabulate(self.df_sort))
#      else:
#        logger.info(df2tabulate(self.df_sort.head(n=5)))
#        logger.info("...")
#        logger.info(df2tabulate(self.df_sort.tail(n=5)))
#        logger.info("")
#        logger.info(colored("Table originally with %i rows is truncated for top and bottom 5 only."%self.df_sort.shape[0], "cyan"))
#        logger.info(colored("Consider filtering it with --n=x for the 1st x results or --filter-tags=foo using a value from your own EC2 tags.", "cyan"))

    if self.n!=0:
      logger.info(colored("This table has been filtered for only the 1st %i underused results"%self.n, "cyan"))




import logging
logger = logging.getLogger('isitfit')

from .tagsSuggestBasic import TagsSuggestBasic
class TagsDump(TagsSuggestBasic):

  def tags_to_dict(self, ec2_obj):
    tags_dict = {x['Key']: x['Value'] for x in ec2_obj.tags}
    return tags_dict


  def suggest(self):
      logger.info("Dumping to csv")
      from .tagsSuggestBasic import dump_df_to_csv
      self.csv_fn = dump_df_to_csv(self.tags_df, 'isitfit-tags-dump-')


  def display(self):
    from .utils import display_df
    display_df(
      "Dumped tags:",
      self.tags_df,
      self.csv_fn,
      self.tags_df.shape,
      logger
    )



import pandas as pd
from tqdm import tqdm
from .pull_cloudtrail_lookupEvents import GeneralManager as GraCloudtrailManager
import os


import logging
logger = logging.getLogger('isitfit')



class Manager:
    def __init__(self, cloudtrail_client, EndTime, cache_man):
        self.cloudtrail_client = cloudtrail_client
        self.EndTime = EndTime
        self.cache_man = cache_man

    def init_data(self, ec2_instances, n_ec2):
        # get cloudtail ec2 type changes for all instances
        self.df_cloudtrail = self._fetch_cached()

        # first pass to append ec2 types to cloudtrail based on "now"
        self.df_cloudtrail = self.df_cloudtrail.reset_index()
        for ec2_obj in tqdm(ec2_instances.all(), total=n_ec2, desc="First pass through EC2 instances", initial=1):
            self._appendNow(ec2_obj)

        # set index again, and sort decreasing this time (not like git-remote-aws default)
        self.df_cloudtrail = self.df_cloudtrail.set_index(["instanceId", "EventTime"]).sort_index(ascending=False)


    def _fetch_cached(self):
        # get cloudtrail ec2 type changes for all instances

        # check cache first
        cache_key = "cloudtrail_ec2type._fetch"
        if self.cache_man.isReady():
          df_cache = self.cache_man.get(cache_key)
          if df_cache is not None:
            logger.debug("Found cloudtrail data in redis cache")
            return df_cache

        # if no cache, then download
        df_fresh = self._fetch_core()

        # if caching enabled, store it for later fetching
        # https://stackoverflow.com/a/57986261/4126114
        if self.cache_man.isReady():
          self.cache_man.set(cache_key, df_fresh)

        # done
        return df_fresh


    def _fetch_core(self):
        # get cloudtrail ec2 type changes for all instances
        logger.debug("Downloading cloudtrail data")
        cloudtrail_manager = GraCloudtrailManager(self.cloudtrail_client)
        df = cloudtrail_manager.ec2_typeChanges()
        return df

    """
    # Cached version ... disabled because not sure how to generalize it
    def _fetch(self):
        # get cloudtail ec2 type changes for all instances
        # FIXME
        cache_fn = '/tmp/isitfit_cloudtrail.shadiakiki1986.csv'
        # cache_fn = '/tmp/isitfit_cloudtrail.autofitcloud.csv'
        if os.path.exists(cache_fn):
            logger.debug("Loading cloudtrail data from cache")
            df = pd.read_csv(cache_fn).set_index(["instanceId", "EventTime"])
            return df

        logger.debug("Downloading cloudtrail data")
        cloudtrail_manager = GraCloudtrailManager(self.cloudtrail_client)
        df = cloudtrail_manager.ec2_typeChanges()

        # save to cache
        df.to_csv(cache_fn)

        # done
        return df
    """



    def _appendNow(self, ec2_obj):
        # artificially append an entry for "now" with the current type
        # This is useful for instance who have no entries in the cloudtrail
        # so that their type still shows up on merge
        self.df_cloudtrail = pd.concat([
            self.df_cloudtrail,
            pd.DataFrame([
              { 'instanceId': ec2_obj.instance_id,
                'EventTime': self.EndTime,
                'instanceType': ec2_obj.instance_type
              }
            ])
          ])


    def single(self, ec2_obj):
        # pandas series of number of cpu's available on the machine over time, past 90 days
        # series_type_ts1 = self.cloudtrail_client.get_ec2_type(ec2_obj.instance_id)
        if not ec2_obj.instance_id in self.df_cloudtrail.index:
            return None

        return self.df_cloudtrail.loc[ec2_obj.instance_id]


